from typing import Sequence

import torch
from torch import nn

from meddlr.metrics.functional.image import mse
from meddlr.metrics.metric import Metric
from meddlr.ops import complex as cplx
from meddlr.utils import env

# Pre-trained model for computing SSFD. This is a UNet model trained on a self-supervsed
# in-painting pretext task on the fastMRI knee dataset.
SSFD_HUGGINGFACE_URL = "https://huggingface.co/philadamson93/SSFD/resolve/main/model.ckpt"


# TODO: Refactor SSFD Class to extract shared logic into parent class FeatureMetric
class SSFD(Metric):
    """
    Self-Supervised Feature Distance. SSFD evaluates the feature distance between a
    pair of images from features extracted from a pre-trained neural network trained on
    a self-supervision task with fastMRI datasets [1]. SSFD has been shown to correspond
    well to Radiologist Reader Scores of accelerated  MR reconstructions.

    References:
    ..  [1] Adamson, Philip M., et al.
        SSFD: Self-Supervised Feature Distance as an MR Image Reconstruction Quality Metric."
        NeurIPS 2021 Workshop on Deep Learning and Inverse Problems. 2021.
        https://openreview.net/forum?id=dgMvTzf6M_3
    """

    is_differentiable = True
    higher_is_better = False

    def __init__(
        self,
        mode: str = "grayscale",
        layer_names: Sequence[str] = ("block4_relu2",),
        channel_names: Sequence[str] = None,
        reduction="none",
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: bool = None,
        dist_sync_fn: bool = None,
    ):
        """
        Args:
            mode (str): Determines how to interpret the channel dimension of the inputs. One of:
                * ``'grayscale'``: Each channel corresponds to a distinct grayscale input image.
                * ``'rgb'``: The 3 channel dimensions correspond to a single rgb image.
                             Exception will be thrown if channel dimension != 3 or dtype is complex
            layer_names (Sequence[str]):
                A list of strings specifying the layers to extract features from. Any of:
                ['block1_relu2', 'block2_relu2', 'block3_relu2', 'block4_relu2', 'block5_relu2']
                SSFD from each layer will be summed if multiple layers are specified.
        """

        super().__init__(
            channel_names=channel_names,
            units="",
            reduction=reduction,
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )

        valid_modes = ("grayscale", "rgb")
        if mode not in valid_modes:
            raise ValueError(f"Invalid `mode` ('{mode}'). Expected one of {valid_modes}.")

        self.mode = mode
        self.layer_names = layer_names

        path_manager = env.get_path_manager()
        file_path = path_manager.get_local_path(SSFD_HUGGINGFACE_URL, force=False)

        self.net = SSFD_Encoder(in_channels=1, out_channels=320)
        self.net.load_state_dict(torch.load(file_path))
        self.net.eval()

    def func(self, preds, targets) -> torch.Tensor:
        if self.mode == "grayscale":
            loss_shape = (targets.shape[0], targets.shape[1])
        elif self.mode == "rgb":
            if targets.shape[1] != 3:
                raise ValueError(
                    f"Channel dimension must have size 3 for rgb mode,\
                    but got tensor of shape {targets.shape}."
                )

            is_complex = cplx.is_complex(targets) or cplx.is_complex_as_real(targets)
            if is_complex:
                raise TypeError(
                    f"Data type must be real when mode is {self.mode},\
                    but got data type {targets.dtype}"
                )

            loss_shape = (targets.shape[0], 1)

        preds = self.preprocess_ssfd(preds)
        targets = self.preprocess_ssfd(targets)

        target_features = self.net(targets)
        pred_features = self.net(preds)

        loss = 0
        for layer in self.layer_names:
            loss += torch.mean(mse(target_features[layer], pred_features[layer]), dim=1)
        loss = loss.view(loss_shape)

        return loss

    def preprocess_ssfd(self, img: torch.Tensor) -> torch.Tensor:
        """
        Preprocess image for SSFD model input.

        Converts to a magnitude scan if complex and normalizes to zero-mean, unit variance.
        If self.mode is 'rgb', then the image will be averaged over the channel dimension.

        Args:
            img (torch.Tensor): Tensor to preprocess.

        Returns:
            img (torch.Tensor): Preprocessed tensor.
        """

        is_complex = cplx.is_complex(img) or cplx.is_complex_as_real(img)
        if is_complex:
            img = cplx.abs(img)

        if self.mode == "grayscale":
            img = img.reshape(img.shape[0] * img.shape[1], 1, img.shape[2], img.shape[3])
        elif self.mode == "rgb":
            img = torch.mean(img, axis=1, keepdim=True)

        shape = (img.shape[0], img.shape[1], -1)

        img_mean = torch.mean(img.reshape(shape), dim=-1, keepdim=True).unsqueeze(-1)
        img_std = torch.std(img.reshape(shape), dim=-1, keepdim=True).unsqueeze(-1)
        img = (img - img_mean) / img_std

        return img


class SSFD_Encoder(nn.Module):
    """
    Pytorch architecture of pre-trained SSFD network.

    This is a UNet architecture with 5 layers and 2 convolutional blocks per layer.
    Modifications to this architecture are not supported as SSFD loads weights from
    a pre-trained network.

    """

    def __init__(self, in_channels: int, out_channels: int):
        """
        Args:
            in_chans (int): Number of channels in the input.
            out_chans (int): Number of channels in the output.
        """

        super().__init__()
        self.in_ch = in_channels
        self.out_ch = out_channels

        self.firstblock = DoubleConv(1, 20)
        self.encoder1 = EncoderBlock(20, 40)
        self.encoder2 = EncoderBlock(40, 80)
        self.encoder3 = EncoderBlock(80, 160)
        self.encoder4 = EncoderBlock(160, 320)

    def forward(self, xin: torch.Tensor):
        """
        Args:
            xin (torch.Tensor): Input tensor of shape [batch_size, self.in_ch, height, width].
        Returns:
            (Dict[str,torch.Tensor]): Dictionary with keys specified by self.layer_names
            and values of the corresponding intermedaite layer output tensors.
        """

        int_layers = {}
        xdown1, int_layers["block1_relu2"] = self.firstblock(xin)
        xdown2, int_layers["block2_relu2"] = self.encoder1(xdown1)
        xdown3, int_layers["block3_relu2"] = self.encoder2(xdown2)
        xdown4, int_layers["block4_relu2"] = self.encoder3(xdown3)
        _, int_layers["block5_relu2"] = self.encoder4(xdown4)

        return int_layers


class DoubleConv(nn.Module):
    """
    A Convolutional Block that consists of two convolution layers each followed by
    group normalization, ReLu activation and dropout.
    """

    def __init__(self, in_ch: int, out_ch: int):
        """
        Args:
            in_ch (int): Number of channels in the input.
            out_ch (int): Number of channels in the output.
        """
        super().__init__()

        self.in_ch = in_ch
        self.out_ch = out_ch

        self.conv1 = nn.Conv2d(in_channels=in_ch, out_channels=out_ch, kernel_size=3, padding=1)
        self.conv2 = nn.Conv2d(in_channels=out_ch, out_channels=out_ch, kernel_size=3, padding=1)
        self.GN = nn.GroupNorm(num_groups=10, num_channels=out_ch)
        self.ReLu = nn.ReLU(inplace=True)
        self.droupout = nn.Dropout(p=0.1)

    def forward(self, x: torch.Tensor):
        """
        Args:
            x (torch.Tensor): Input tensor of shape [batch_size, self.in_ch, height, width].
        Returns:
            x (torch.Tensor): Output tensor of shape [batch_size, self.out_ch, height, width].
            int_layer (torch.Tensor): Intermediate output tensor after the second ReLu activation
                                      of shape [batch_size, self.out_ch, height, width]
        """
        x = self.conv1(x)
        x = self.ReLu(x)
        x = self.conv2(x)
        x = self.ReLu(x)
        int_layer = x  # intermediate layer to extract features from for SSFD
        x = self.GN(x)

        return x, int_layer


class EncoderBlock(nn.Module):
    """
    A Convolutional Block that consists of Max Pooling followed by two convolution layers
    each followed by group normalization, ReLu activation and dropout.
    """

    def __init__(self, in_ch: int, out_ch: int):
        """
        Args:
            in_ch (int): Number of channels in the input.
            out_ch (int): Number of channels in the output.
        """
        super().__init__()
        self.MP = nn.MaxPool2d(2)
        self.conv_block = DoubleConv(in_ch, out_ch)

    def forward(self, x: torch.Tensor):
        """
        Args:
            x (torch.Tensor): Input tensor of shape [batch_size, self.in_ch, height, width].
        Returns:
            x (torch.Tensor): Output tensor of shape [batch_size, self.out_ch, height, width].
            int_layer (torch.Tensor): Intermediate output tensor from self.conv_block
                                      of shape [batch_size, self.out_ch, height, width]
        """
        x = self.MP(x)
        x, int_layer = self.conv_block(x)

        return x, int_layer



import inspect
import logging
from copy import deepcopy
from typing import Any, Dict

from meddlr.metrics.collection import MetricCollection
from meddlr.metrics.image import MAE, MSE, NRMSE, PSNR, RMSE, SSIM
from meddlr.metrics.lpip import LPIPS
from meddlr.metrics.sem_seg import ASSD, CV, DSC, VOE
from meddlr.metrics.ssfd import SSFD

logger = logging.getLogger(__name__)


# The builtin metrics that will be recognized by every
# evaluator based on the string completion.
# We specify the class and any keyword args required to
# construct the object.
# fmt: off
_BUILTIN_METRICS = {
    # =========== Image =========== #
    # PSNR
    "psnr": (PSNR, {"im_type": None}),
    "psnr_mag": (PSNR, {"im_type": "magnitude"}),
    "psnr_real": (PSNR, {"im_type": "real"}),
    "psnr_phase": (PSNR, {"im_type": "phase"}),
    "psnr_imag": (PSNR, {"im_type": "imag"}),
    # NRMSE
    "nrmse": (NRMSE, {"im_type": None}),
    "nrmse_mag": (NRMSE, {"im_type": "magnitude"}),
    "nrmse_real": (NRMSE, {"im_type": "real"}),
    "nrmse_phase": (NRMSE, {"im_type": "phase"}),
    "nrmse_imag": (NRMSE, {"im_type": "imag"}),
    # MSE
    "mse": (MSE, {"im_type": None}),
    "mse_mag": (MSE, {"im_type": "magnitude"}),
    "mse_real": (MSE, {"im_type": "real"}),
    "mse_phase": (MSE, {"im_type": "phase"}),
    "mse_imag": (MSE, {"im_type": "imag"}),
    # RMSE
    "rmse": (RMSE, {"im_type": None}),
    "rmse_mag": (RMSE, {"im_type": "magnitude"}),
    "rmse_real": (RMSE, {"im_type": "real"}),
    "rmse_phase": (RMSE, {"im_type": "phase"}),
    "rmse_imag": (RMSE, {"im_type": "imag"}),
    # SSIM
    "ssim (Wang)": (SSIM, {"method": "wang"}),
    # MAE
    "mae": (MAE, {"im_type": None}),
    "mae_mag": (MAE, {"im_type": "magnitude"}),
    "mae_real": (MAE, {"im_type": "real"}),
    "mae_phase": (MAE, {"im_type": "phase"}),
    "mae_imag": (MAE, {"im_type": "imag"}),

    # =========== Semantic Segmentation =========== #
    "DSC": (DSC, {}),
    "VOE": (VOE, {}),
    "ASSD": (ASSD, {"connectivity": 1}),
    "CV": (CV, {}),

    # ======== Feature Based Metrics ============= #
    # Defaults of net_type: 'alex', lpips: True and pretrained: True chosen based on LPIPS paper:
    #   R. Zhang, P. Isola, A. A. Efros, E. Shechtman, O. Wang.
    #   The Unreasonable Effectiveness of Deep Features as a Perceptual Metric.
    #   In CVPR, 2018 http://arxiv.org/abs/1801.03924
    # and the LPIPS Github Repo: https://github.com/richzhang/PerceptualSimilarity
    # Default of mode: grayscale chosen as we work with grayscale images in MRI.
    "LPIPS": (LPIPS, {"net_type": "alex",
                      "mode": "grayscale",
                      "lpips": True,
                      "pretrained": True
                      }),
    # Default of layer_names: ('block4_relu2') chosen based on original SSFD implementation:
    #   Adamson, Philip M., et al.
    #   SSFD: Self-Supervised Feature Distance as an MR Image Reconstruction Quality Metric."
    #   NeurIPS 2021 Workshop on Deep Learning and Inverse Problems. 2021.
    #   https://openreview.net/forum?id=dgMvTzf6M_3
    # Default of mode: grayscale chosen as we work with grayscale images in MRI.
    "SSFD": (SSFD, {"mode": "grayscale",
                    "layer_names": ("block4_relu2",)})
}
# fmt: on


def build_metrics(metric_names, fmt: str = None, **kwargs) -> MetricCollection:
    metrics = {}
    for name in metric_names:
        if name not in _BUILTIN_METRICS.keys():
            raise ValueError(
                "Metric '{}' not found in built-in metrics. "
                "Built-in metrics include:\n{}".format(name, _BUILTIN_METRICS.keys())
            )

        klass: type
        base_kwargs: Dict[str, Any]
        klass, base_kwargs = _BUILTIN_METRICS[name]
        signature = inspect.signature(klass)
        klass_kwargs = {k: v for k, v in kwargs.items() if k in signature.parameters}

        shared_keys = base_kwargs.keys() & klass_kwargs.keys()
        if shared_keys:
            logger.warning(
                "Metric '{}' default values for following arguments will be overridden. "
                "This is may lead to non-reproducible results:\n{}".format(name, shared_keys)
            )
        base_kwargs = deepcopy(base_kwargs)
        base_kwargs.update(klass_kwargs)

        mname = fmt.format(name) if fmt else name
        metrics[mname] = klass(**base_kwargs)

    return MetricCollection(metrics)



import logging
import os
import pathlib
import shutil

import torch

from meddlr.config import CfgNode
from meddlr.evaluation.testing import find_weights
from meddlr.utils import env
from meddlr.utils.general import move_to_device

logger = logging.getLogger(__name__)


def prepare_model(
    cfg: CfgNode,
    name: str,
    output_dir: str,
    weights_file: str = None,
    criterion: str = None,
    dry_run: bool = False,
    device: str = "cpu",
) -> pathlib.Path:
    """Processes an experiment folder for model deployment.

    This function finds the model checkpoint (based on the val_loss criterion)
    and config files in the experiment folder. If the model checkpoint is on the
    GPU, it is moved to the CPU. The model is then saved to the output directory.

    If your model training was performed with some accelerator (e.g. GPU),
    this function must be run on a node with that accelerator.

    Important fields to set in the config:
        - DESCRIPTION.BRIEF: A brief description of the model.
        - DESCRIPTION.TAGS: A list of tags to describe the model.
        - OUTPUT_DIR: The output directory for users who want to use
          the config for training. This should start with ``results://``.

    Note:
        This function is in pre-alpha stage. There may be several breaking changes
        to this function.

    Examples:
        >>> prepare_model(cfg, "my_model", "shareable-models", criterion="val_ssim")

    Returns:
        os.PathLike: The path to the directory containing the formatted model and config.
    """
    path = cfg.OUTPUT_DIR
    # project = cfg.DESCRIPTION.PROJECT_NAME
    cfg = cfg.clone().defrost()

    path_manager = env.get_path_manager()

    logger.info(f"================== {name} ==================")
    cfg.OUTPUT_DIR = path_manager.get_local_path(path)
    path_manager.get_local_path(os.path.join(path, "metrics.csv"))

    if weights_file is None:
        weights_file, _, _ = find_weights(cfg, criterion=criterion)

    fmt_name = name.replace(" ", "_").replace("(", "").replace(")", "").replace(",", "_")
    exp_out_dir = os.path.join(output_dir, fmt_name)

    if dry_run:
        logger.info(f"{name}: {weights_file} -> {exp_out_dir}")
        return

    if os.path.exists(exp_out_dir):
        shutil.rmtree(exp_out_dir)
    os.makedirs(exp_out_dir, exist_ok=True)

    weights = torch.load(weights_file)
    weights = move_to_device(weights, device=device)
    torch.save(weights, os.path.join(exp_out_dir, "model.ckpt"))

    # Config
    out_cfg_file = f"{exp_out_dir}/config.yaml"
    cfg.DESCRIPTION.EXP_NAME = name
    cfg.DESCRIPTION.ENTITY_NAME = ""
    cfg.DESCRIPTION.PROJECT_NAME = ""
    cfg.MODEL.DEVICE = device
    # exp_type = project if project in exp_name.lower() else "baseline"
    # cfg.DESCRIPTION.TAGS = (
    #     exp_type,
    #     exp_name.split(" ")[0].lower(),
    #     str(cfg.AUG_TRAIN.UNDERSAMPLE.ACCELERATIONS[0]) + "x",
    #     dataset.split("/")[0],
    # )
    cfg.TEST.VAL_METRICS.RECON = [
        metric for metric in cfg.TEST.VAL_METRICS.RECON if "ssim_old" not in metric
    ]
    with open(out_cfg_file, "w") as f:
        f.write(cfg.dump())
    return pathlib.Path(exp_out_dir)



from typing import Sequence

import torch

from meddlr.metrics.metric import Metric
from meddlr.ops import complex as cplx
from meddlr.utils import env

if env.package_available("lpips"):
    from lpips import LPIPS as _LPIPS


# TODO: Refactor SSFD Class to extract shared logic into parent class FeatureMetric
class LPIPS(Metric):
    """
    Learned Perceptual Image Patch Similarity.

    LPIPS evaluates the feature distance between a pair of images from features extracted
    from a pre-trained neural network [1]. LPIPS has been shown to correspond well to
    perceived image quality on natural images.

    References:
    ..  [1] R. Zhang, P. Isola, A. A. Efros, E. Shechtman, O. Wang.
        The Unreasonable Effectiveness of Deep Features as a Perceptual Metric.
        In CVPR, 2018 http://arxiv.org/abs/1801.03924
    """

    is_differentiable = True
    higher_is_better = False

    def __init__(
        self,
        net_type: str = "alex",
        mode: str = "grayscale",
        lpips: bool = True,
        pretrained: bool = True,
        channel_names: Sequence[str] = None,
        reduction="none",
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: bool = None,
        dist_sync_fn: bool = None,
    ):
        """
        Args:
        net_type (str): The pre-trained network to use for extracting features. One of:
            * ``'alex'``: Alex-Net w/ feature extraction layers 'relu1' through 'relu5'
            * ``'vgg'``: VGG-16 w/ feature extration layers ['relu1_2', 'relu2_2',
                       'relu3_3', 'relu4_3', 'relu5_3']
            * ``'squeeze'``: Squeeze-Net w/ feature extration layers 'relu1' through 'relu7'
        mode (str): Determines how to interpret the channel dimension of the inputs. One of:
            * ``'grayscale'``: Each channel corresponds to a distinct grayscale input image.
            * ``'rgb'``: The 3 channel dimensions correspond to a single rgb image.
                         Exception will be thrown if channel dimension != 3 dtype data is complex.
        lpips (bool): This flag determines if a linear layer is used on top of the
                      extracted features.
            * ``True``: linear layers on top of base/trunk network.
            * ``False``: no linear layers; each layer is averaged together.
        pretrained (bool): This flag controls the linear layers, which are only in
                           effect when lpips=True above.
            * ``True``: linear layers are calibrated with human perceptual judgments.
            * ``False``: linear layers are randomly initialized.
        """

        if not env.package_available("lpips"):
            raise ModuleNotFoundError(
                "LPIPS metric requires that lpips is installed."
                "Either install as `pip install meddlr[metrics]` or `pip install lpips`."
            )

        super().__init__(
            channel_names=channel_names,
            units="",
            reduction=reduction,
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )

        valid_net_type = ("vgg", "alex", "squeeze")
        if net_type not in valid_net_type:
            raise ValueError(
                f"Invalid `net_type` ('{net_type}'). Expected one of {valid_net_type}."
            )

        valid_modes = ("grayscale", "rgb")
        if mode not in valid_modes:
            raise ValueError(f"Invalid `mode` ('{mode}'). Expected one of {valid_modes}.")

        self.net = NoTrainLpips(net=net_type, lpips=lpips, verbose=False)
        self.mode = mode

    def func(self, preds: torch.Tensor, targets: torch.Tensor):

        if self.mode == "grayscale":
            loss_shape = (targets.shape[0], targets.shape[1])
        elif self.mode == "rgb":
            if targets.shape[1] != 3:
                raise ValueError(
                    f"Channel dimension must have size 3 for rgb mode,\
                    but got tensor of shape {targets.shape}."
                )

            is_complex = cplx.is_complex(targets) or cplx.is_complex_as_real(targets)
            if is_complex:
                raise TypeError(
                    f"Data type must be real when mode is {self.mode},\
                    but got data type {targets.dtype}"
                )

            loss_shape = (targets.shape[0], 1)

        preds = self.preprocess_lpips(preds)
        targets = self.preprocess_lpips(targets)

        loss = self.net(preds, targets).squeeze()
        loss = loss.view(loss_shape)

        return loss

    def preprocess_lpips(self, img: torch.Tensor) -> torch.Tensor:
        """
        Preprocess image per LPIPS implementation.

        Converts images to magnitude images if complex and normalizes between [-1, 1].
        If self.mode is 'grayscale', then each channel dimension will be replicated 3 times.

        Args:
            img (torch.Tensor): Tensor to preprocess.

        Returns:
            img (torch.Tensor): Preprocessed tensor.
        """

        is_complex = cplx.is_complex(img) or cplx.is_complex_as_real(img)
        if is_complex:
            img = cplx.abs(img)

        if self.mode == "grayscale":
            # normalize each image independently (channel dim. represents different images)
            shape = (img.shape[0], img.shape[1], -1)
            img_min = torch.amin(img.reshape(shape), dim=-1, keepdim=True).unsqueeze(-1)
            img_max = torch.amax(img.reshape(shape), dim=-1, keepdim=True).unsqueeze(-1)
            img = 2 * (img - img_min) / (img_max - img_min) - 1

            img = img.reshape(img.shape[0] * img.shape[1], 1, img.shape[2], img.shape[3])
            img = img.repeat(1, 3, 1, 1)
        elif self.mode == "rgb":
            # normalize each image independently (channel dim. represents the same image)
            shape = (img.shape[0], -1)
            img_min = (
                torch.amin(img.reshape(shape), dim=-1, keepdim=True).unsqueeze(-1).unsqueeze(-1)
            )
            img_max = (
                torch.amax(img.reshape(shape), dim=-1, keepdim=True).unsqueeze(-1).unsqueeze(-1)
            )
            img = 2 * (img - img_min) / (img_max - img_min) - 1

        return img


if env.package_available("lpips"):

    class NoTrainLpips(_LPIPS):
        def train(self, mode: bool) -> "NoTrainLpips":
            """the network should not be able to be switched away from evaluation mode.
            Implementation adapted from torchmetrics LPIPS."""
            return super().train(False)

else:
    NoTrainLpips = None



from typing import Optional, Tuple

import torch

from meddlr.ops import complex as cplx
from meddlr.transforms.build import TRANSFORM_REGISTRY
from meddlr.transforms.transform import Transform


@TRANSFORM_REGISTRY.register()
class NoiseTransform(Transform):
    """A deterministic transform that adds zero-mean complex additive Gaussian (white) noise.

    Note:
        This transform currently only supports adding noise to k-space.
        It does not support adding noise to image data.

    TODO (arjundd): Add support for adding noise to arbitrary data types.
    """

    def __init__(
        self,
        std_dev: float,
        use_mask: bool = True,
        rho: float = None,
        seed: int = None,
        generator: torch.Generator = None,
    ):
        """
        Args:
            std_dev: The standard deviation of the noise.
            use_mask: Whether to only add noise to the non-zero entries of the kspace.
            rho: The fraction of entries to add noise to. If ``use_mask=True``,
                only non-zero entries are considered.
            seed (int, optional): The seed to use for the random number generator.
            generator (torch.Generator, optional): The random number generator to use.
                Must be specified if ``seed`` is not set.
        """
        if seed is None and generator is None:
            raise ValueError("One of `seed` or `generator` must be specified.")
        self.std_dev = std_dev
        self.use_mask = use_mask
        self.rho = rho
        self.seed = seed

        gen_state = None
        if generator is not None:
            gen_state = generator.get_state()
        self._generator_state = gen_state

    def apply_kspace(self, kspace: torch.Tensor) -> torch.Tensor:
        """Add noise to the kspace.

        Args:
            kspace: A complex valued tensor.

        Returns:
            torch.Tensor: A complex valued tensor with noise added.
        """
        return self._add_noise(kspace)

    def _generator(self, data: torch.Tensor) -> torch.Generator:
        seed = self.seed

        g = torch.Generator(device=data.device)
        if seed is None:
            g.set_state(self._generator_state)
        else:
            g = g.manual_seed(seed)
        return g

    def _add_noise(self, data: torch.Tensor) -> torch.Tensor:
        noise_std = self.std_dev
        subsample_masks = self.rho is not None and self.rho != 1

        mask = None
        if self.use_mask:
            mask = cplx.get_mask(data)
        if mask is None and subsample_masks:
            mask = torch.ones(data.shape)

        g = self._generator(data)
        if cplx.is_complex(data):
            noise = noise_std * torch.randn(data.shape + (2,), generator=g, device=data.device)
            noise = torch.view_as_complex(noise)
        else:
            noise = noise_std * torch.randn(data.shape, generator=g, device=data.device)

        if self.rho is not None and self.rho != 1:
            mask = self.subsample_mask(mask)
        if mask is not None:
            noise = noise * mask
        aug_kspace = data + noise

        return aug_kspace

    def subsample_mask(
        self, mask: torch.Tensor, generator: Optional[torch.Generator] = None
    ) -> torch.Tensor:
        """Returns mask with ``1-self.rho`` fraction of valid entries dropped.

        Valid entries in ``mask`` are non-zero entries. Only non-zero entries will be masked out.

        TODO (arjundd): Use :cls:`KspaceMaskTransform` for this.

        Args:
            mask: A binary tensor where ``1`` indicates valid entries.

        Returns:
            torch.Tensor: The subsampled mask.
        """
        rho = self.rho
        shape = mask.shape
        mask = mask.view(-1)

        # TODO: this doesnt work if the matrix is > 2*24 in size.
        # TODO: make this a bit more optimized
        num_valid = torch.sum(mask)
        weights = mask / num_valid
        num_samples = int((1 - rho) * num_valid)
        samples = torch.multinomial(weights, num_samples, replacement=False, generator=generator)
        mask[samples] = 0

        mask = mask.view(shape)
        return mask

    def _eq_attrs(self) -> Tuple[str]:
        return ("std_dev", "use_mask", "rho", "seed", "_generator_state")



from typing import Dict, Optional

import torch
import torchvision.utils as tv_utils
from torch import nn

from meddlr.config.config import configurable
from meddlr.data.transforms.noise import NoiseModel
from meddlr.modeling.meta_arch.build import META_ARCH_REGISTRY, build_model
from meddlr.modeling.meta_arch.ssdu import SSDUModel
from meddlr.ops import complex as cplx
from meddlr.utils.events import get_event_storage


@META_ARCH_REGISTRY.register()
class N2RModel(nn.Module):
    """Noise2Recon model.

    Reference:
        AD Desai, BM Ozturkler, CM Sandino, et al. Noise2Recon: A Semi-Supervised Framework
        for Joint MRI Reconstruction and Denoising. ArXiv 2021.
        https://arxiv.org/abs/2110.00075
    """

    _version = 3

    @configurable
    def __init__(
        self,
        model: nn.Module,
        noiser: NoiseModel,
        use_supervised_consistency: bool = False,
        vis_period: int = -1,
    ):
        """
        Args:
            model: The base model.
            noiser: The additive noise module.
            use_supervised_consistency: Whether to apply noise-based consistency
                to supervised examples.
            vis_period (int, optional): The period over which to visualize images.
                If ``<=0``, it is ignored. Note if the ``model`` has a ``vis_period``
                attribute, it will be overridden so that this class handles visualization.
        """
        super().__init__()
        self.model = model

        # Visualization done by this model.
        # If sub-model is SSDU, we allow SSDU to log images
        # for the recon pipeline.
        if (
            not isinstance(self.model, SSDUModel)
            and hasattr(self.model, "vis_period")
            and vis_period > 0
        ):
            self.model.vis_period = -1
        self.vis_period = vis_period

        # Whether to keep gradient for base images in transform.
        self.use_base_grad = False
        # Use supervised examples for consistency
        self.use_supervised_consistency = use_supervised_consistency
        self.noiser = noiser

    def augment(self, inputs):
        """Noise augmentation module for the consistency branch.

        Args:
            inputs (Dict[str, Any]): The input dictionary.
                It must contain a key ``'kspace'``, which traditionally
                corresponds to the undersampled kspace when performing
                augmentation for consistency.

        Returns:
            Dict[str, Any]: The input dictionary with the kspace polluted
                with additive masked complex Gaussian noise.
        """
        kspace = inputs["kspace"].clone()
        aug_kspace = self.noiser(kspace, clone=False)

        inputs = {k: v.clone() for k, v in inputs.items() if k != "kspace"}
        inputs["kspace"] = aug_kspace
        return inputs

    @torch.no_grad()
    def visualize_aug_training(self, kspace, kspace_aug, preds, preds_base, target=None):
        """Visualize training of augmented data.

        Args:
            kspace: The base kspace.
            kspace_aug: The augmented kspace.
            preds: Reconstruction of augmented kspace. Shape: NxHxWx2.
            preds_base: Reconstruction of base kspace. Shape: NxHxWx2.
        """
        storage = get_event_storage()

        # calc mask for first coil only
        if cplx.is_complex(kspace):
            kspace = torch.view_as_real(kspace)
        kspace = kspace.cpu()[0, ..., 0, :].unsqueeze(0)
        if cplx.is_complex(kspace_aug):
            kspace_aug = torch.view_as_real(kspace_aug)
        kspace_aug = kspace_aug.cpu()[0, ..., 0, :].unsqueeze(0)
        preds = preds.cpu()[0, ...].unsqueeze(0)
        preds_base = preds_base.cpu()[0, ...].unsqueeze(0)

        all_images = [preds, preds_base]
        errors = [cplx.abs(preds_base - preds)]
        if target is not None:
            target = target.cpu()[0, ...].unsqueeze(0)
            all_images.append(target)
            errors.append(cplx.abs(target - preds))

        all_images = torch.cat(all_images, dim=2)
        all_kspace = torch.cat([kspace, kspace_aug], dim=2)
        errors = torch.cat(errors, dim=2)

        imgs_to_write = {
            "phases": cplx.angle(all_images),
            "images": cplx.abs(all_images),
            "errors": errors,
            "masks": cplx.get_mask(kspace),
            "kspace": cplx.abs(all_kspace),
        }

        for name, data in imgs_to_write.items():
            data = data.squeeze(-1).unsqueeze(1)
            data = tv_utils.make_grid(data, nrow=1, padding=1, normalize=True, scale_each=True)
            storage.put_image("train_aug/{}".format(name), data.numpy(), data_format="CHW")

    def _format_consistency_inputs(
        self,
        inputs_supervised: Optional[Dict[str, torch.Tensor]] = None,
        inputs_unsupervised: Optional[Dict[str, torch.Tensor]] = None,
    ) -> Dict[str, torch.Tensor]:
        """Generate base and augmented inputs to be used for consistency training.

        Args:
            inputs_supervised: A dict of inputs, their metadata, and their ground truth references.
            inputs_unsupervised: A dict of inputs and their metadata.

        Returns:
            Dict[str, Dict[str, Tensor]]: A dictionary of base inputs and augmented inputs:
                - 'base': Inputs to be used to generate the pseudo-label (i.e. target)
                  for consistency optimization.
                - 'aug': Noise augmented inputs to use for consistency training.
        """
        inputs_consistency = []
        if inputs_unsupervised is not None:
            inputs_consistency.append(inputs_unsupervised)
        if self.use_supervised_consistency and inputs_supervised is not None:
            inputs_consistency.append({k: v for k, v in inputs_supervised.items() if k != "target"})

        if len(inputs_consistency) == 0:
            return {}  # No consistency training.

        if len(inputs_consistency) > 1:
            inputs_consistency = {
                k: torch.cat([x[k] for x in inputs_consistency], dim=0)
                for k in inputs_consistency[0].keys()
            }
        else:
            inputs_consistency = inputs_consistency[0]

        # Augment the inputs.
        inputs_consistency_aug = self.augment(inputs_consistency)

        return {"base": inputs_consistency, "aug": inputs_consistency_aug}

    def forward(self, inputs):
        if not self.training:
            assert (
                "unsupervised" not in inputs
            ), "unsupervised inputs should not be provided in eval mode"
            inputs = inputs.get("supervised", inputs)
            return self.model(inputs)

        storage = get_event_storage()
        vis_training = self.training and self.vis_period > 0 and storage.iter % self.vis_period == 0

        inputs_supervised = inputs.get("supervised", None)
        inputs_unsupervised = inputs.get("unsupervised", None)
        if inputs_supervised is None and inputs_unsupervised is None:
            raise ValueError("Examples not formatted in the proper way")
        # Whether to use self-supervised via data undersampling (SSDU) for reconstruction.
        is_ssdu_enabled = isinstance(self.model, SSDUModel)
        output_dict = {}

        # Reconstruction (supervised).
        if is_ssdu_enabled:
            output_dict["recon"] = self.model(
                inputs,
            )
        elif inputs_supervised is not None:
            output_dict["recon"] = self.model(
                inputs_supervised, return_pp=True, vis_training=vis_training
            )

        # Consistency (unsupervised).
        # kspace_aug = kspace + U \sigma \mathcal{N}
        # Loss = L(f(kspace_aug, \theta), f(kspace, \theta))
        # If the model is an SSDU model, unpack it to use the internal model for consistency.
        model = self.model.model if is_ssdu_enabled else self.model
        consistency_inputs = self._format_consistency_inputs(inputs_supervised, inputs_unsupervised)
        if len(consistency_inputs) > 0:
            inputs_consistency = consistency_inputs["base"]
            inputs_consistency_aug = consistency_inputs["aug"]

            with torch.no_grad():
                pred_base = model(inputs_consistency)
                # Target only used for visualization purposes not for loss.
                target = inputs_consistency.get("target", None)
                pred_base = pred_base["pred"]

            pred_aug: Dict[str, torch.Tensor] = model(inputs_consistency_aug, return_pp=True)

            pred_aug.pop("target", None)
            pred_aug["target"] = pred_base.detach()
            output_dict["consistency"] = pred_aug

            if vis_training:
                self.visualize_aug_training(
                    inputs_consistency["kspace"],
                    inputs_consistency_aug["kspace"],
                    pred_aug["pred"],
                    pred_base,
                    target=target,
                )

        return output_dict

    def load_state_dict(self, state_dict, strict=True):  # pragma: no cover
        # TODO: Configure backwards compatibility
        if any(x.startswith("unrolled") for x in state_dict.keys()):
            raise ValueError(
                "`self.unrolled` was renamed to `self.model`. "
                "Backwards compatibility has not been configured."
            )
        return super().load_state_dict(state_dict, strict)

    @classmethod
    def from_config(cls, cfg):
        model_cfg = cfg.clone()
        model_cfg.defrost()
        model_cfg.MODEL.META_ARCHITECTURE = cfg.MODEL.N2R.META_ARCHITECTURE
        model_cfg.freeze()
        model = build_model(model_cfg)

        noiser = NoiseModel.from_cfg(cfg)

        return {
            "model": model,
            "noiser": noiser,
            "use_supervised_consistency": cfg.MODEL.N2R.USE_SUPERVISED_CONSISTENCY,
            "vis_period": cfg.VIS_PERIOD,
        }



import os
import warnings
from typing import Any, Callable, Dict, Union

import torch

import meddlr.config.util as config_util
from meddlr.config import CfgNode, get_cfg
from meddlr.modeling.meta_arch import build_model
from meddlr.utils import env

__all__ = ["get_model_from_zoo"]


def get_model_from_zoo(
    cfg_or_file: Union[str, os.PathLike, CfgNode],
    weights_path: Union[str, os.PathLike] = None,
    strict: bool = True,
    ignore_shape_mismatch: bool = False,
    force_download: bool = False,
    build_model_fn: Callable = None,
    build_model_kwargs: Dict[str, Any] = None,
) -> torch.nn.Module:
    """Get model from zoo and optionally load in weights.

    This function is designed for distributing models for use.
    It builds the model from a configuration and optionally loads in pre-trained weights.

    Pre-trained weights can either be specified by the ``weights_file`` argument
    or by setting the config option ``cfg.MODEL.WEIGHTS``. If neither is specified,
    the model is initialized randomly. If ``weights_file`` is an empty string,
    ``cfg.MODEL.WEIGHTS`` will also be ignored, and the model will be initialized randomly.

    Args:
        cfg_or_file (str | path-like | CfgNode): The config (or file).
            If it is a file, it will be merged with the default config.
        weights_path (str | path-like, optional): The weights file to load.
            This can also be specified in the ``cfg.MODEL.WEIGHTS`` config
            field. If neither are provided, the uninitialized model will be returned.
        strict (bool, optional): Strict option to pass to ``load_state_dict``.
        ignore_shape_mismatch (bool, optional): If ``True``, weights that do not
            match the model layer's shape will be ignored instead of raising
            an error.
        force_download (bool, optional): Force download of model config/weights
            if downloading the model.
        build_model_fn (callable, optional): A function to build the model.
            Defaults to :func:`meddlr.modeling.meta_arch.build_model`.
        build_model_kwargs (dict, optional): Keyword arguments to pass to
            ``build_model_fn``.

    Returns:
        torch.nn.Module: The model loaded with pre-trained weights.

    Examples:
        .. code-block:: python
            from meddlr.config import get_cfg
            from meddlr.engine.model_zoo import get_model_from_zoo

            # Load a pretrained model from a config file.
            cfg_file = "https://huggingface.co/arjundd/vortex-release/raw/main/mridata_knee_3dfse/Supervised/config.yaml"  # noqa: E501
            weights_path = "https://huggingface.co/arjundd/vortex-release/resolve/main/mridata_knee_3dfse/Supervised/model.ckpt"  # noqa: E501
            model = get_model_from_zoo(cfg_file, weights_path)

            # Build a model from config file with randomly initialized weights.
            model = get_model_from_zoo(cfg_file, weights_path="")

            # Build model from a config object.
            cfg = get_cfg()
            cfg.MODEL.META_ARCHITECTURE = "GeneralizedUnrolledCNN"
            model = get_model_from_zoo(cfg)
    """
    path_manager = env.get_path_manager()

    if not isinstance(cfg_or_file, CfgNode):
        cfg_file = path_manager.get_local_path(cfg_or_file, force=force_download)
        failed_deps = config_util.check_dependencies(cfg_file, return_failed_deps=True)
        if len(failed_deps) > 0:  # pragma: no cover
            warning_msg = (
                f"Some dependenices are not met. "
                f"This may result in some issues with model construction or weight loading. "
                f"Unmet dependencies: {failed_deps}"
            )
            warnings.warn(warning_msg)
        cfg = get_cfg()
        cfg.merge_from_file(cfg_file)
    else:
        cfg = cfg_or_file

    if build_model_fn is None:
        build_model_fn = build_model
    if not build_model_kwargs:
        build_model_kwargs = {}
    model = build_model_fn(cfg, **build_model_kwargs)

    if weights_path is None:
        weights_path = cfg.MODEL.WEIGHTS
    if not weights_path:
        return model

    weights_path = path_manager.get_local_path(weights_path, force=force_download)
    model = load_weights(
        model, weights_path, strict=strict, ignore_shape_mismatch=ignore_shape_mismatch
    )
    return model


def load_weights(
    model: torch.nn.Module,
    weights_path: Union[str, os.PathLike],
    strict: bool = True,
    ignore_shape_mismatch: bool = False,
    force_download: bool = False,
    find_device: bool = True,
) -> torch.nn.Module:
    """Load model from checkpoint.

    This function is designed for distributing models for use. It loads in pre-trained weights.
    Pre-trained weights can either be specified by the ``weights_file`` argument
    or by setting the config option ``cfg.MODEL.WEIGHTS``. If neither is specified,
    the model is initialized randomly. If ``weights_file`` is an empty string,
    ``cfg.MODEL.WEIGHTS`` will also be ignored, and the model will be initialized randomly.

    Args:
        model (torch.nn.Module): The model to load weights into.
        checkpoint_path (str | path-like): The checkpoint file to load.
        strict (bool, optional): Strict option to pass to ``load_state_dict``.
        ignore_shape_mismatch (bool, optional): If ``True``, weights that do not
            match the model layer's shape will be ignored instead of raising
            an error.
        force_download (bool, optional): Force download of model config/weights
            if downloading the model.
        find_device (bool, optional): If ``True``, find the device the weights
            were stored on and moves the model prior to loading weights.

    Returns:
        torch.nn.Module: The model loaded with pre-trained weights.
            Note this model will be unpacked from the PyTorch Lightnining modules.
    """
    path_manager = env.get_path_manager()

    weights_path = path_manager.get_local_path(weights_path, force=force_download)
    weights = torch.load(weights_path)
    if "state_dict" in weights:
        weights = weights["state_dict"]
    elif "model" in weights:
        weights = weights["model"]

    if ignore_shape_mismatch:
        params_shape_mismatch = _find_mismatch_sizes(model, weights)
        if len(params_shape_mismatch) > 0:
            mismatched_params_str = "".join("\t- {}\n".format(x) for x in params_shape_mismatch)
            warnings.warn(
                "Shape mismatch found for some parameters. Ignoring these weights:\n{}".format(
                    mismatched_params_str
                )
            )
            for p in params_shape_mismatch:
                weights.pop(p)
            strict = False

    if find_device:
        device = next(iter(weights.values())).device
        model = model.to(device)

    model.load_state_dict(weights, strict=strict)

    return model


def _find_mismatch_sizes(model: torch.nn.Module, state_dict):
    """Finds the keys in the state_dict that are different from the model.
    Args:
        model (torch.nn.Module): The model to load weights into.
        state_dict (dict): The state_dict to load.
    Returns:
        list[str]: The list of keys that are different from the model.
    """
    params_shape_mismatch = set()
    for source in [model.state_dict().items(), model.named_parameters()]:
        for name, param in source:
            if name not in state_dict:
                continue
            if param.shape != state_dict[name].shape:
                params_shape_mismatch |= {name}

    return list(params_shape_mismatch)



import copy
import itertools
import logging
import os
import time
import uuid

import torch
from tqdm import tqdm

import meddlr.ops as oF
import meddlr.utils.comm as comm
from meddlr.data.transforms.transform import build_normalizer
from meddlr.evaluation.scan_evaluator import ScanEvaluator, structure_scans
from meddlr.metrics.build import build_metrics

_DEFAULT_METRICS = (
    "DSC",
    "VOE",
    "CV",
    "DSC_scan",
    "VOE_scan",
    "CV_scan",
    "ASSD_scan",
)


class SemSegEvaluator(ScanEvaluator):
    """
    Evaluate semantic segmentation quality.
    """

    def __init__(
        self,
        dataset_name,
        cfg,
        distributed=False,
        sync_outputs=False,
        aggregate_scans=True,
        output_dir=None,
        group_by_scan: bool = False,
        save_seg: bool = False,
        metrics=None,
        flush_period: bool = None,
        to_cpu=True,
    ):
        """
        Args:
            dataset_name (str): name of the dataset to be evaluated.
            cfg (CfgNode): config instance
            output_dir (str): optional, an output directory to dump all
                results predicted on the dataset. The dump contains two files:

                1. "instance_predictions.pth" a file in torch serialization
                   format that contains all the raw original predictions.
                2. "coco_instances_results.json" a json file in COCO's result
                   format.
            distributed (bool, optional): If `True`, collect results from all
                ranks for evaluation.
                Otherwise, will evaluate the results in the current process.
                If using `DistributedDataParallel`, this should likely be `True`.
            sync_outputs (bool, optional): If `True`, synchronizes all predictions
                before evaluation.
                If `False`, synchronizes metrics before reduction.
                Ignored if `distributed=False`
            aggregate_scans (bool, optional): If `True`, also computes metrics per
                scan under the label `'scan_{metric}'` (e.g. scan_l1).

        Not yet supported:
            running_eval (bool, optional): If `True`, evaluates scans while processing.
                This reduces the memory overhead. Note if distributed=`True`, sync_outputs
                must be `False`.
        """
        self._output_dir = output_dir
        self._dataset_name = dataset_name
        self._cpu_device = torch.device("cpu")
        self._logger = logging.getLogger(__name__)
        self._normalizer = build_normalizer(cfg)
        self._distributed = distributed
        self._sync_outputs = sync_outputs
        self._aggregate_scans = aggregate_scans
        self._group_by_scan = group_by_scan
        self._save_seg = save_seg
        self.to_cpu = to_cpu
        self._is_flushing = False

        self.activation = cfg.MODEL.SEG.ACTIVATION
        self.threshold = 0.5
        # if self._distributed and self._sync_outputs and running_eval:
        #     raise ValueError("running_eval not possible when outputs must be synced")

        self.has_bg = cfg.MODEL.SEG.INCLUDE_BACKGROUND or cfg.MODEL.SEG.ACTIVATION == "softmax"
        if self.has_bg:
            # Add "background" to class names if desired.
            self._class_names = (
                cfg.MODEL.SEG.CLASSES
                if any(x.lower() in ["background", "bg"] for x in cfg.MODEL.SEG.CLASSES)
                else ("bg",) + tuple(cfg.MODEL.SEG.CLASSES)
            )
        else:
            self._class_names = cfg.MODEL.SEG.CLASSES

        self._metric_names = []
        if metrics is not False:
            if not metrics:
                metrics = self.default_metrics()
            self._metric_names = metrics

        if flush_period is None:
            flush_period = cfg.TEST.FLUSH_PERIOD
        self.flush_period = flush_period

        self._remaining_state = None
        self._predictions = []
        self._is_flushing = False

    @classmethod
    def default_metrics(cls):
        """The default metrics processed by this class."""
        # TODO: Add support for ASSD_scan
        return _DEFAULT_METRICS

    def reset(self):
        self._predictions = []
        self._remaining_state = None
        self._is_flushing = False

        metrics = self._metric_names
        slice_metrics = [m for m in metrics if not m.endswith("_scan")]
        scan_metrics = [m[:-5] for m in metrics if m.endswith("_scan")]

        self.slice_metrics = build_metrics(slice_metrics, channel_names=self._class_names)
        self.scan_metrics = build_metrics(
            scan_metrics, fmt="{}_scan", channel_names=self._class_names
        )
        self.slice_metrics.eval()
        self.scan_metrics.eval()

    def structure_scans(self, verbose=True):
        """Structure scans into volumes to be used to evaluation."""
        out = structure_scans(self._predictions, verbose=verbose, dims={1: "slice_id"})
        return out

    def process(self, inputs, outputs):
        """
        Args:
            inputs: the inputs to a recon model (e.g., GeneralizedRCNN).
                Currently this should be an empty dictionary.
            outputs: the outputs of a COCO model. It is a list of dicts with key
                "instances" that contains :class:`Instances`.
        """
        labels: torch.Tensor = self._parse_sem_seg_label(inputs)
        pred: torch.Tensor = self._parse_sem_seg_pred(outputs)

        N = pred.shape[0]
        preds = pred.type(torch.bool)
        targets = labels.type(torch.bool)
        if self.to_cpu:
            preds = preds.to(self._cpu_device, non_blocking=True)
            targets = targets.to(self._cpu_device, non_blocking=True)

        self._predictions.extend(
            [
                {
                    "pred": preds[i],
                    "target": targets[i],
                    "metadata": inputs["metadata"][i] if "metadata" in inputs else {},
                }
                for i in range(N)
            ]
        )

        has_num_examples = self.flush_period > 0 and len(self._predictions) >= self.flush_period
        has_num_scans = self.flush_period < 0 and len(
            {x["metadata"]["scan_id"] for x in self._predictions}
        ) > abs(self.flush_period)
        if has_num_examples or has_num_scans:
            self.flush(enter_prediction_scope=True, skip_last_scan=True)

    def synchronize_predictions(self):
        comm.synchronize()
        self._predictions = comm.gather(self._predictions, dst=0)
        self._predictions = list(itertools.chain(*self._predictions))
        if not comm.is_main_process():
            self._predictions = []

    def evaluate(self):
        # Sync predictions (if applicable)
        if self._distributed and self._sync_outputs:
            self.synchronize_predictions()
            if not self._predictions:
                return {}

        if len(self._predictions) == 0:
            self._logger.warning(f"[{type(self).__name__}] Did not receive valid predictions.")
            return {}

        # TODO: Add support in metrics manager for appending metrics to the same scan_id
        for _idx, pred in enumerate(tqdm(self._predictions, desc="Slice Metrics")):
            if "metadata" in pred:
                scan_id = "/".join(
                    str(pred["metadata"].get(k, uuid.uuid4())) for k in ["scan_id", "slice_id"]
                )
                voxel_spacing = pred["metadata"].get("voxel_spacing", None)
                if voxel_spacing:
                    voxel_spacing = voxel_spacing[-(pred["pred"].ndim - 1) :]
            else:
                scan_id = str(uuid.uuid4())
                voxel_spacing = None
            self.slice_metrics(
                ids=[scan_id],
                preds=pred["pred"].unsqueeze(0),
                targets=pred["target"].unsqueeze(0),
                voxel_spacing=voxel_spacing,
                spacing=voxel_spacing,
            )

        # Compute metrics per scan.
        if self._aggregate_scans:
            scans = self.structure_scans()
            for scan_id, pred in tqdm(scans.items(), desc="Scan Metrics"):
                self.scan_metrics(
                    ids=[scan_id],
                    preds=pred["pred"].unsqueeze(0),
                    targets=pred["target"].unsqueeze(0),
                    category_dim=0,
                    voxel_spacing=pred.get("voxel_spacing", None),
                    spacing=pred.get("voxel_spacing", None),
                )

        pred_vals = self.slice_metrics.to_dict()
        pred_vals.update(self.scan_metrics.to_dict())
        self._results = pred_vals

        if not self._is_flushing:
            self.log_summary()

        # Copy so the caller can do whatever with results
        return copy.deepcopy(self._results)

    def log_summary(self):
        slice_metrics_summary = self.slice_metrics.summary()
        scan_metrics_summary = self.scan_metrics.summary()
        slice_metrics_df = self.slice_metrics.to_pandas()
        scan_metrics_df = self.scan_metrics.to_pandas()

        if not comm.is_main_process():
            return

        output_dir = self._output_dir
        self._logger.info(
            "[{}] Slice metrics summary:\n{}".format(type(self).__name__, slice_metrics_summary)
        )
        self._logger.info(
            "[{}] Scan metrics summary:\n{}".format(type(self).__name__, scan_metrics_summary)
        )

        if not output_dir:
            return

        dirpath = output_dir
        os.makedirs(dirpath, exist_ok=True)
        test_results_summary_path = os.path.join(dirpath, "results.txt")
        slice_metrics_path = os.path.join(dirpath, "slice_metrics.csv")
        scan_metrics_path = os.path.join(dirpath, "scan_metrics.csv")

        # Write details to test file
        with open(test_results_summary_path, "w+") as f:
            f.write("Results generated on %s\n" % time.strftime("%X %x %Z"))
            # f.write("Weights Loaded: %s\n" % os.path.basename(self._config.TEST_WEIGHT_PATH))

            f.write("--" * 40)
            f.write("\n")
            f.write("Slice Metrics:\n")
            f.write(slice_metrics_summary)
            f.write("--" * 40)
            f.write("\n")
            f.write("Scan Metrics:\n")
            f.write(scan_metrics_summary)
            f.write("--" * 40)
            f.write("\n")

        slice_metrics_df.to_csv(slice_metrics_path, header=True, index=True)
        scan_metrics_df.to_csv(scan_metrics_path, header=True, index=True)

    def _parse_sem_seg_pred(self, input):
        # TODO: Update these values
        num_classes = 1
        channel_dim = 1

        keys = ["sem_seg_pred", "sem_seg_probs", "sem_seg_logits", "pred", "probs", "logits"]
        existing_keys = [k for k in keys if k in input.keys()]
        if len(existing_keys) == 0:
            raise ValueError(f"`input` must have one of keys: {keys}.\n\tGot: {list(input.keys())}")
        key = existing_keys[0]
        assert input[key].shape[1] == len(self._class_names), input[key].shape

        if key.endswith("pred"):
            pred = input[key]
        elif key.endswith("logits"):
            logits = input[key]
            if self.activation == "sigmoid":
                probs = torch.sigmoid(logits)
                pred = (probs >= self.threshold).type(torch.long)
            elif self.activation == "softmax":
                pred = oF.pred_to_categorical(
                    logits, activation=self.activation, channel_dim=1, threshold=0.5
                )
                pred = oF.to_onehot(pred, len(num_classes) + 1)
        elif key.endswith("probs"):
            probs = input[key]
            if self.activation == "sigmoid":
                pred = probs >= self.threshold
            elif self.activation == "softmax":
                pred = oF.to_onehot(
                    torch.argmax(probs, dim=channel_dim),
                    num_classes=len(self._class_names) + 1,
                )

        return pred.type(torch.uint8)

    def _parse_sem_seg_label(self, output):
        keys = ["sem_seg", "sem_seg_labels", "sem_seg_target", "labels", "target"]
        existing_keys = [k for k in keys if k in output.keys()]
        if len(existing_keys) == 0:
            raise ValueError(
                f"`output` must have one of keys: {keys}.\n\tGot: {list(output.keys())}"
            )
        key = existing_keys[0]
        assert output[key].shape[1] == len(self._class_names), f"Key '{key}': {output[key].shape}"
        return output[key].type(torch.uint8)



import json
import logging
import os
import pprint
import re
from numbers import Number
from typing import Any, Dict, List, Mapping, Tuple, Union

import numpy as np
import pandas as pd
import torch

from meddlr.config import CfgNode


def print_csv_format(results: Dict[str, Dict[str, Number]]):  # pragma: no cover
    """Print metrics for easy copypaste.

    Args:
        results (OrderedDict[dict]): task_name -> {metric -> score}
    """
    assert isinstance(results, Dict), results  # unordered results cannot be properly printed
    logger = logging.getLogger(__name__)
    important_res = [(k, v) for k, v in results.items() if "-" not in k]
    logger.info("copypaste: " + ",".join([k[0] for k in important_res]))
    logger.info("copypaste: " + ",".join(["{0:.4f}".format(k[1]) for k in important_res]))

    # Additional formatting
    logger.info(
        "Metrics (comma delimited): \n{}\n{}".format(
            ",".join([k[0] for k in important_res]),
            ",".join(["{0:.4f}".format(k[1]) for k in important_res]),
        )
    )


def verify_results(cfg: CfgNode, results: Dict[str, Any]) -> bool:  # pragma: no cover
    """Verify that the results are consistent with what is expected in the config.

    Adapted from detectron2:
    https://github.com/facebookresearch/detectron2/blob/main/detectron2/evaluation/testing.py

    Args:
        results: A mapping from metrics -> scores.

    Returns:
        bool: Whether the verification succeeds or not
    """
    expected_results = cfg.TEST.EXPECTED_RESULTS
    if not len(expected_results):
        return True

    results = flatten_results_dict(results)

    ok = True
    for metric, expected, tolerance in expected_results:
        actual = results[metric]
        if not np.isfinite(actual):
            ok = False
        diff = abs(actual - expected)
        if diff > tolerance:
            ok = False

    logger = logging.getLogger(__name__)
    if not ok:
        logger.error("Result verification failed!")
        logger.error("Expected Results: " + str(expected_results))
        logger.error("Actual Results: " + pprint.pformat(results))
    else:
        logger.info("Results verification passed.")
    return ok


def flatten_results_dict(results: Dict[str, Any], delimiter: str = "/") -> Dict[str, Number]:
    """
    Expand a hierarchical dict of scalars into a flat dict of scalars.
    If results[k1][k2][k3] = v, the returned dict will have the entry
    {"k1/k2/k3": v}.

    Args:
        results (dict):
    """
    r = {}
    for k, v in results.items():
        if isinstance(v, Mapping):
            v = flatten_results_dict(v)
            for kk, vv in v.items():
                r[k + delimiter + kk] = vv
        else:
            r[k] = v
    return r


# Supported validation metrics and the operation to perform on them.
# TODO (arjundd): Do not hardcode these. Match to metrics in meddlr instead.
_METRICS_TO_OPERATION = {
    "l1": "min",
    "l2": "min",
    "psnr": "max",
    "ssim": "max",
    "iteration": "max",  # find the last checkpoint
    "loss": "min",
    "mae": "min",
}


def find_metrics(
    cfg: CfgNode, criterion: str, operation: str = "auto", iter_limit: int = None, top_k: int = 1
):
    """Find the iteration(s) resulting in best metrics for a given criterion.

    Args:
        cfg: The config.
        criterion: The criterion that we use to select weights.
        operation: The operation for selecting the best value of the criterion.
            One of `'auto'`, `'min'`, `'max'`.
        iter_limit: If specified, only iterations ``<=iter_limit` will be searched.
            If this value is negative, it is interpreted as the epoch limit.
            A positive value is recommended for avoiding ambiguity.
        top_k: The number of (iteration, metric) pairs to return.
            If ``-1``, return all pairs in sorted in order of best to worst.

    Returns:
        List[Tuple[int, float]] | Tuple[int, float]: A list of tuples of the
            form (iteration, metric) if ``top_k > 1``. Otherwise, a tuple of
            the form (iteration, metric).

    Note:
        This function is experimental. The API may change without warning.
    """
    metric_details = _find_metric_details(
        cfg=cfg,
        criterion=criterion,
        operation=operation,
        iter_limit=iter_limit,
        top_k=-1,
    )
    best_iter_and_values = metric_details["best_iter_and_values"][:top_k]
    return best_iter_and_values[0] if top_k == 1 else best_iter_and_values


def find_weights(
    cfg: CfgNode,
    criterion: str,
    operation: str = "auto",
    iter_limit: int = None,
    file_name_fmt: str = "model_{:07d}.pth",
    top_k: int = 1,
) -> Tuple[Union[str, List[str]], str, Union[float, List[float]]]:
    """Find the best weights based on a metric criterion.

    TODO (arjundd): Use :func:`find_metrics` to find the appropriate metrics.

    Args:
        cfg: The config.
        criterion (str): The criterion that we use to select weights.
        operation (str, optional): The operation for the best value of the criterion.
            One of `'auto'`, `'min'`, `'max'`.
        iter_limit (int, optional): If specified, all weights will be before
            this iteration. If this value is negative, it is
            interpreted as the epoch limit.
        file_name_fmt (int, optional): The naming format for checkpoint files.
        top_k (int, optional): The number of top checkpoints to return.

    Returns:
        Tuple: ``k`` filepath(s), selection criterion, and ``k`` criterion value(s).
            If ``k=1``, filepath is a string and value is a float.

    Examples:
        >>> checkpoint_file, criterion, value = find_weights(cfg, "val_loss", "min")

    Note:
        This function is experimental. The API may change without warning.
    """
    logger = logging.getLogger(__name__)

    metric_details = _find_metric_details(
        cfg=cfg,
        criterion=criterion,
        operation=operation,
        iter_limit=iter_limit,
        top_k=-1,
    )
    best_iter_and_values = metric_details["best_iter_and_values"][:top_k]
    last_iter = metric_details["last_iter"]

    all_filepaths = []
    all_values = []
    potential_ckpt_files = os.listdir(cfg.OUTPUT_DIR)
    for best_iter, best_value in best_iter_and_values:
        file_name = file_name_fmt.format(best_iter)

        matched_files = [x for x in potential_ckpt_files if re.match(file_name, x)]
        if len(matched_files) > 1:
            raise ValueError(
                f"Too many potential checkpoint files found for iter={best_iter}, "
                f"criterion={criterion}, value={best_value}:\n\t{matched_files}"
            )
        if len(matched_files) == 0:
            if best_iter == last_iter:
                file_name = "model_final.pth"
            matched_files = [file_name for x in potential_ckpt_files if re.match(file_name, x)]
        if len(matched_files) == 0:
            raise ValueError(
                f"Could not find potential checkpoint files for iter={best_iter}, "
                f"criterion={criterion}, value={best_value}."
            )
        file_name = matched_files[0]

        file_path = os.path.join(cfg.OUTPUT_DIR, file_name)

        if not os.path.isfile(file_path):
            raise ValueError("Model for iteration {} does not exist".format(best_iter))

        all_filepaths.append(file_path)
        all_values.append(best_value)

        logger.info("Weights: {} - {}: {:0.4f}".format(file_name, criterion, best_value))

    if top_k == 1:
        return all_filepaths[0], criterion, all_values[0]
    else:
        return all_filepaths, criterion, all_values


def check_consistency(state_dict: Dict[str, Any], model: torch.nn.Module):
    """Verifies that the proper weights were loaded into the model.

    Related to issue that loading weights from checkpoints of a cuda model
    does not properly load when model is on cpu. It may also result in
    warnings for contiguous tensors, but this is not always the case.
    https://github.com/pytorch/pytorch/issues/42300

    Args:
        state_dict (Dict): A model state dict.
        model (nn.Module): A Pytorch model.
    """
    _state_dict = model.state_dict()
    for k in state_dict:
        assert k in _state_dict, f"{k} not in model state_dict: {_state_dict.keys()}"
        assert torch.equal(state_dict[k], _state_dict[k]), f"Mismatch values: {k}"


def _metrics_from_x(metrics_file, criterion):
    metrics_file = os.path.splitext(metrics_file)[0]
    if os.path.isfile(f"{metrics_file}.json"):
        return _metrics_from_json(f"{metrics_file}.json", criterion)
    elif os.path.isfile(f"{metrics_file}.csv"):
        return _metrics_from_csv(f"{metrics_file}.csv", criterion)
    else:
        raise ValueError(f"metrics file not found - {metrics_file}")


def _metrics_from_json(metrics_file, criterion):
    metrics = []
    with open(metrics_file, "r") as f:
        metrics = [json.loads(line.strip()) for line in f]
    metrics = [m for m in metrics if criterion in m or any(k.endswith(criterion) for k in m.keys())]
    is_metric_wrapped = criterion not in metrics[0]
    if is_metric_wrapped:
        metrics = [
            (
                int(m["iteration"]),
                np.mean(
                    [m[k] for k in m if k.endswith(criterion) and "test" not in k.split("/")[0]]
                ).item(),
            )
            for m in metrics
        ]
    else:
        metrics = [(m["iteration"], m[criterion]) for m in metrics]
    return metrics


def _metrics_from_csv(metrics_file, criterion):
    def _row_has_criterion(row: pd.Series):
        index = row.index.tolist()
        return criterion in index and not pd.isna(row[criterion])

    metrics = pd.read_csv(metrics_file)
    metrics = [m for _, m in metrics.iterrows() if _row_has_criterion(m)]
    metrics = [(int(m["step"]), m[criterion]) for m in metrics]
    return metrics


def get_iters_per_epoch_eval(cfg) -> int:
    """Get number of iterations per epoch for evaluation purposes.

    This function expects a metrics named ``{cfg.OUTPUT_DIR}/metrics.json``
    that is written during training.

    Args:
        cfg: The config.

    Return:
        int: Number of iterations per epoch.
    """
    exp_path = cfg.OUTPUT_DIR

    eval_period = cfg.TEST.EVAL_PERIOD
    time_scale = cfg.TIME_SCALE
    assert time_scale in ["epoch", "iter"]
    if (time_scale == "epoch" and eval_period < 0) or (time_scale == "iter" and eval_period > 0):
        raise ValueError("Evaluation period is not in # epochs")
    eval_period = abs(eval_period)

    with open(os.path.join(exp_path, "metrics.json"), "r") as f:
        metrics = [json.loads(line.strip()) for line in f.readlines()]
    # Filter to only have evaluation metrics.
    metrics = [m for m in metrics if "eval_time" in m]
    # last eval could be at end of training which doesnt have same
    # spacing of iters_per_epoch - skip it
    iterations = np.diff([m["iteration"] for m in metrics][:-1])
    if len(iterations) == 0:
        raise ValueError("Could not determine iters_per_epoch - too few evaluations")
    if len(iterations) >= 2 and not np.all(iterations[:-1] == iterations[0]):
        raise ValueError("Not all iteration spacings are equal - {iterations}")

    iters_per_epoch = iterations[0] / eval_period
    assert iters_per_epoch == int(iters_per_epoch)
    return int(iters_per_epoch)


def _find_metric_details(
    cfg: CfgNode,
    criterion: str,
    operation: str = "auto",
    iter_limit: int = None,
    top_k: int = 1,
):
    """Find the iteration(s) resulting in best metrics for a given criterion.

    Args:
        cfg: The config.
        criterion: The criterion that we use to select weights.
        operation: The operation for selecting the best value of the criterion.
            One of `'auto'`, `'min'`, `'max'`.
        iter_limit: If specified, only iterations ``<=iter_limit` will be searched.
            If this value is negative, it is interpreted as the epoch limit.
            A positive value is recommended for avoiding ambiguity.
        top_k: The number of (iteration, metric) pairs to return.
            If ``-1``, return all pairs in sorted in order of best to worst.

    Returns:
        List[Tuple[int, float]] | Tuple[int, float]: A list of tuples of the
            form (iteration, metric) if ``top_k > 1``. Otherwise, a tuple of
            the form (iteration, metric). Also returns last iteration if requested.

    Note:
        This function is experimental. The API may change without warning.
    """
    logger = logging.getLogger(__name__)

    if operation not in ["min", "max", "auto"]:
        raise ValueError(f"Invalid operation: {operation}. Expected one of 'min', 'max', 'auto'.")

    # Negative iter_limit is interpreted as epoch limit.
    if iter_limit is not None and iter_limit < 0:
        iter_limit = int(abs(iter_limit) * get_iters_per_epoch_eval(cfg))

    ckpt_period = cfg.SOLVER.CHECKPOINT_PERIOD
    eval_period = cfg.TEST.EVAL_PERIOD
    if (
        ckpt_period * eval_period <= 0  # same sign (i.e. same time scale)
        or abs(eval_period) % abs(ckpt_period) != 0  # checkpoint period is multiple of eval period
    ):
        raise ValueError(  # pragma: no cover
            "Cannot find weights if checkpoint/eval periods "
            "at different time scales or eval period is not "
            "a multiple of checkpoint period."
        )

    if operation == "auto":
        operation = [
            op for name, op in _METRICS_TO_OPERATION.items() if name.lower() in criterion.lower()
        ]
        if len(operation) == 0:
            raise ValueError(f"Could not find operation for criterion '{criterion}'.")
        if len(operation) > 1:
            raise ValueError(f"Found multiple operations for criterion '{criterion}': {operation}")
        operation = operation[0]
    assert operation in ["min", "max"]

    logger.info("Finding best weights in {} using {}...".format(cfg.OUTPUT_DIR, criterion))

    # Filter metrics to find reporting of real validation metrics.
    # If metric is wrapped (e.g. "mridata_knee_2019_val/val_l1"), that means
    # multiple datasets were validated on.
    # We filter out metrics from datasets that contain the word "test".
    # The criterion from all other datasets are averaged and used as the
    # target criterion.
    metrics_file = os.path.join(cfg.OUTPUT_DIR, "metrics")
    try:
        metrics = _metrics_from_x(metrics_file, criterion)
    except IndexError:
        raise ValueError(f"No metrics found matching criterion '{criterion}'.")

    # Last iteration that was logged.
    # Note if the run did not complete, `model_final.pth` may not correspond
    # with this iteration. Use with caution.
    last_iter = metrics[-1][0]

    # Filter out all metrics calculated above iter limit.
    if iter_limit:
        metrics = [x for x in metrics if x[0] < iter_limit]

    # Retraining does not overwrite the metrics file.
    # We make sure that the metrics correspond only to the most
    # recent run.
    metrics_recent_order = metrics[::-1]
    iterations = [m[0] for m in metrics_recent_order]
    intervals = np.diff(iterations)
    if any(intervals > 0):
        stop_idx = np.argmax(intervals > 0) + 1
        metrics_recent_order = metrics_recent_order[:stop_idx]
        metrics = metrics_recent_order[::-1]

    # Note that resuming can sometimes report metrics for the same
    # iteration. We handle this by taking the most recent metric for the
    # iteration __after__ filtering out old training runs.
    # metrics = {iteration: value for iteration, value in metrics}
    metrics = [(iteration, value) for iteration, value in metrics]

    best_iter_and_values = sorted(metrics, key=lambda x: x[1], reverse=operation == "max")
    if top_k > 0:
        best_iter_and_values = best_iter_and_values[:top_k]
    return {
        "best_iter_and_values": best_iter_and_values,
        "last_iter": last_iter,
    }



import copy
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from fvcore.common.registry import Registry
from torch import nn

CUSTOM_LAYERS_REGISTRY = Registry("CUSTOM_LAYERS")
CUSTOM_LAYERS_REGISTRY.__doc__ = """
Registry for custom layers.

Use this registry to identify if the layer is not provided by default in torch.nn.
"""

_LAYER_SHORTCUTS = {
    "bn": "batchnorm",
    "dropout1d": "dropout",
}
_PT_LAYERS_LOWERCASE = {
    name.lower(): layer
    for name, layer in nn.__dict__.items()
    if isinstance(layer, type) and issubclass(layer, nn.Module)
}

__all__ = ["get_layer_type", "get_layer_kind"]


def get_layer_type(name: str, dimension: int = None) -> type:
    """Returns the layer type based on the name and, in some cases, the dimension.

    This function searches both the default PyTorch layers and the custom layers
    registered in :obj:`CUSTOM_LAYERS_REGISTRY`.

    Args:
        name (str): Name of the layer.
        dimension (int, optional): Dimension of the layer.
            Note, not all layers require this argument (e.g. ReLU), but
            it may be safe to pass it regardless.

    Returns:
        type: Layer type.

    Raises:
        ValueError: If the layer type is not found with the name/dimension pair.
    """
    in_name = name
    name = name.lower()

    # Handle some layer names earlier because of naming conventions.
    if name == "dropout" and dimension:
        name += f"{dimension}d"

    if name in _LAYER_SHORTCUTS:
        name = _LAYER_SHORTCUTS[name]

    custom_layer_names = {x.lower(): x for x in CUSTOM_LAYERS_REGISTRY._obj_map}
    if name in CUSTOM_LAYERS_REGISTRY:
        return CUSTOM_LAYERS_REGISTRY.get(name)
    if name in custom_layer_names:
        return CUSTOM_LAYERS_REGISTRY.get(custom_layer_names[name])
    if name in _PT_LAYERS_LOWERCASE:
        return _PT_LAYERS_LOWERCASE.get(name)

    if any(x in name for x in ["norm", "conv"]):
        if not dimension:
            raise ValueError(f"{in_name} requires dimension")

    if dimension:
        name += f"{dimension}d"
    if name in CUSTOM_LAYERS_REGISTRY:
        return CUSTOM_LAYERS_REGISTRY.get(name)
    if name in custom_layer_names:
        return CUSTOM_LAYERS_REGISTRY.get(custom_layer_names[name])
    if name in _PT_LAYERS_LOWERCASE:
        return _PT_LAYERS_LOWERCASE.get(name)

    raise ValueError(f"No layer found for '{in_name}'")


def get_layer_kind(layer_type: Union[type, str]) -> str:
    """Returns the layer kind based on the layer type.

    The layer kind is a string that describes the kind of layer:

        - "conv": Convolutional layer.
        - "norm": Normalization layer.
        - "act": Activation layer.
        - "dropout": Dropout layer.
        - "unknown": Unknown layer.

    This delineation is useful for building models that order layers based
    on their kind. For example, a model with layers ``conv->norm->act->dropout``
    would need to know the kind of the different types of the layers to organize
    them appropriately.

    TODO: Add support for pooling layers.

    Args:
        layer_type (Union[type, str]): Layer type.

    Returns:
        str: Layer kind.
    """
    if isinstance(layer_type, str):
        layer_type = get_layer_type(layer_type, dimension=2)
    assert issubclass(layer_type, nn.Module)
    kinds_in_name = ["norm", "conv"]
    name = layer_type.__name__
    lower_name = name.lower()
    for kind in kinds_in_name:
        if kind in lower_name:
            return kind

    if hasattr(nn.modules.activation, name):
        return "act"

    if "dropout" in lower_name:
        return "dropout"

    return "unknown"


_LayerInfoInitKwargsDict = Dict[str, Any]
_LayerInfoInitKwargsFlatSequence = Union[List[Any], Tuple[Any]]
_NestedInnerArgs = Tuple[str, Any]
_LayerInfoInitKwargsNestedSequence = Union[List[_NestedInnerArgs], Tuple[_NestedInnerArgs, ...]]
_LayerInfoInitKwargsType = Union[
    _LayerInfoInitKwargsDict, _LayerInfoInitKwargsFlatSequence, _LayerInfoInitKwargsNestedSequence
]

# LayerInfo type schema with raw Python types (str, int, float, tuple, list, etc.).
LayerInfoRawType = Union[
    str, Dict[str, _LayerInfoInitKwargsType], Tuple[str, _LayerInfoInitKwargsType]
]


@dataclass
class LayerInfo:
    """Dataclass for managing layer information."""

    name: str  # name of the layer
    dimension: Optional[int] = None  # The dimension of the layer.
    init_kwargs: Dict[str, Any] = field(default_factory=dict)  # keyword args to initialize layer

    @property
    def kind(self) -> str:
        """The layer kind."""
        return get_layer_kind(self.ltype)

    @property
    def ltype(self) -> type:
        """The layer type."""
        return get_layer_type(self.name, dimension=self.dimension)

    @classmethod
    def format(cls, layer_info: LayerInfoRawType) -> "LayerInfo":
        """Formats layer information from Python raw types to LayerInfo object.

        Args:
            layer_info:

        Returns:
            LayerInfo
        """
        if isinstance(layer_info, str):
            return LayerInfo(name=layer_info, init_kwargs={})

        if isinstance(layer_info, Dict):
            if len(layer_info) != 1:
                raise ValueError(
                    "Dictionary format for LayerInfo can only have one key-value pair - "
                    "e.g. {'dropout': {'p': 0.5}}. "
                    f"Got {layer_info}"
                )
            name, init_kwargs = list(layer_info.items())[0]
        elif isinstance(layer_info, (Tuple, List)):
            if len(layer_info) != 2:
                raise ValueError(
                    "Sequence format for LayerInfo should be formatted as (name, init_kwargs) - "
                    "e.g. ('dropout': {'p': 0.5}), ('dropout', ('p', 0.5). "
                    f"Got {layer_info}"
                )
            name, init_kwargs = layer_info[0], layer_info[1]
        else:
            raise ValueError(f"Unsupported layer info format: {type(layer_info)}")

        # Format init kwargs.
        init_kwargs_err_message = (
            "Unknown init_kwargs format. init_kwargs must follow one of these formats: "
            "\n\t- dict: {key1: value1, key2: value2}"
            "\n\t- flat sequence: [key1, value1, key2, value2, ...]"
            "\n\t- nested sequence: [(key1, value1), (key2, value2), ...]\n"
            f"Got {init_kwargs}"
        )
        if isinstance(init_kwargs, (Tuple, List)):
            if all(isinstance(x, (Tuple, List)) and len(x) == 2 for x in init_kwargs):
                # Nested sequence - i.e. [(key1, value1), (key2, value2), ...]
                init_kwargs = {x[0]: x[1] for x in init_kwargs}
            else:
                if len(init_kwargs) % 2 != 0:
                    raise ValueError(init_kwargs_err_message)
                # Flat sequence - i.e. [key1, value1, key2, value2, ...]
                init_kwargs = dict(zip(init_kwargs[::2], init_kwargs[1::2]))
        if not isinstance(init_kwargs, Dict):
            raise ValueError(init_kwargs_err_message)

        return LayerInfo(name=name, init_kwargs=init_kwargs)

    def build(self, *args, **kwargs) -> nn.Module:
        """Builds the layer.

        Args:
            *args: Positional arguments for building the module.
            **kwargs: Keyword arguments to pass to the layer's ``__init__``.
                Note, these will override the init_kwargs of the layer if there
                are conflicts.
        """
        return self.ltype(*args, **{**self.init_kwargs, **kwargs})


def build_layer_info_from_seq(
    layers_info: Sequence[Union[LayerInfoRawType, LayerInfo]], dimension: Optional[int] = None
) -> List[LayerInfo]:
    """Builds a sequence of layer info objects from a sequence of layer info types.

    Args:
        layers_info: Sequence of layer info types.

    Returns:
        List[LayerInfo]: list of LayerInfo objects corresponding to each element
            in layers_info.
    """
    out = [
        copy.deepcopy(x) if isinstance(x, LayerInfo) else LayerInfo.format(x) for x in layers_info
    ]
    if dimension is not None:
        # TODO (arjundd): We probably want some extra logic here where we don't
        # the dimension if it already exists in the name (e.g. conv1d).
        # Currently, if the name contains the dimension and a dimension is passed
        # (e.g. conv1d, 2), then the dimension value is ignored.
        for x in out:
            x.dimension = dimension

    return out



import inspect
import os
from typing import List, Sequence, Tuple, Union

import numba as nb
import numpy as np
import sigpy.mri
import torch
import torch.nn.functional as F
from fvcore.common.registry import Registry

import meddlr.ops.complex as cplx

MASK_FUNC_REGISTRY = Registry("MASK_FUNC")
MASK_FUNC_REGISTRY.__doc__ = """
Registry for mask functions, which create undersampling masks of a specified
shape.
"""


def build_mask_func(cfg, **kwargs):
    name = cfg.UNDERSAMPLE.NAME
    accelerations = cfg.UNDERSAMPLE.ACCELERATIONS
    calibration_size = cfg.UNDERSAMPLE.CALIBRATION_SIZE
    center_fractions = cfg.UNDERSAMPLE.CENTER_FRACTIONS

    klass = MASK_FUNC_REGISTRY.get(name)
    parameters = inspect.signature(klass).parameters

    # Optional args
    kwargs = kwargs.copy()
    mapping = {"max_attempts": cfg.UNDERSAMPLE.MAX_ATTEMPTS}
    for param, value in mapping.items():
        if param in parameters:
            kwargs[param] = value

    return klass(accelerations, calibration_size, center_fractions, **kwargs)


class MaskFunc:
    """Abstract MaskFunc class for creating undersampling masks of a specified
    shape.

    Adapted from Facebook fastMRI.
    """

    def __init__(self, accelerations):
        """
        Args:
            accelerations (List[int]): Range of acceleration rates to simulate.
        """
        self.accelerations = accelerations
        self.rng = np.random.RandomState()

    def choose_acceleration(self):
        """Chooses a random acceleration rate given a range.

        If self.accelerations is a constant, it will be returned

        """
        if not isinstance(self.accelerations, Sequence):
            return self.accelerations
        elif len(self.accelerations) == 1:
            return self.accelerations[0]
        accel_range = self.accelerations[1] - self.accelerations[0]
        acceleration = self.accelerations[0] + accel_range * self.rng.rand()
        return acceleration

    def get_edge_mask(self, kspace: torch.Tensor, out_shape: Sequence[int] = None) -> torch.Tensor:
        """Get the mask of the edges of kspace that are not sampled.

        ``True`` values indicate the point is an edge location.

        To accelerate the scan, edges of kspace are often not sampled
        or are zero-padded during reconstruction. This method returns
        an estimate of these edges based on the kspace.

        Different undersampling methods have different mechanisms for estimating
        the edges.

        This method should be applied on the fully-sampled kspace (when possible)
        for most accurate edge estimation. It should also be applied prior to
        any additional padding on the kspace (e.g. ZIP2). If padding is used,
        the edge mask must be manually padded as well.

        Args:
            kspace (torch.Tensor): The kspace to estimate the edges of.
                Shape depends on the implementation in the subclass.

        Returns:
            torch.Tensor: A mask of the edges of kspace.
            out_shape: The optional shape of the output mask.
                There will be some constraints on the shape based on the
                subclass implementation. Most typically, it will require
                that the spatial dimensions of the output_shape match
                the spatial dimensions of the kspace.
        """
        raise NotImplementedError()


class CacheableMaskMixin:
    def get_filename(self):
        raise NotImplementedError


@MASK_FUNC_REGISTRY.register()
class RandomMaskFunc(MaskFunc):
    """
    RandomMaskFunc creates a 2D uniformly random undersampling mask.
    """

    def __init__(self, accelerations, calib_size, center_fractions=None):
        if center_fractions:
            raise ValueError(f"center_fractions not yet supported for class {type(self)}.")
        super().__init__(accelerations)
        self.calib_size = calib_size

    def __call__(self, out_shape, seed=None, acceleration=None):
        # Design parameters for mask
        nky = out_shape[1]
        nkz = out_shape[2]

        if not acceleration:
            acceleration = self.choose_acceleration()
        prob = 1.0 / acceleration

        # Generate undersampling mask.
        rand_kwargs = {"dtype": torch.float32}
        if seed is not None:
            rand_kwargs["generator"] = torch.Generator().manual_seed(seed)

        mask = torch.rand([nky, nkz], **rand_kwargs)
        mask = torch.where(mask < prob, torch.Tensor([1]), torch.Tensor([0]))

        # Add calibration region
        calib = [self.calib_size, self.calib_size]
        mask[
            int(nky / 2 - calib[-2] / 2) : int(nky / 2 + calib[-2] / 2),
            int(nkz / 2 - calib[-1] / 2) : int(nkz / 2 + calib[-1] / 2),
        ] = torch.Tensor([1])

        return mask.reshape(out_shape)

    def get_edge_mask(self, kspace: torch.Tensor, out_shape: Sequence[int] = None):
        """See :method:`MaskFunc.get_edge_mask`.

        Expected `kspace` shape (batch, ky, kz, ...).
        """
        # TODO: dims should be configured based on the number of dimenions in the input.
        return get_cartesian_edge_mask(kspace, dims=(1, 2), out_shape=out_shape)


@MASK_FUNC_REGISTRY.register()
class PoissonDiskMaskFunc(CacheableMaskMixin, MaskFunc):
    """
    PoissonDiskMaskFunc creates a 2D Poisson disk undersampling mask.
    """

    def __init__(
        self,
        accelerations: Union[int, Tuple[int, int]],
        calib_size: Union[int, Tuple[int, int]],
        center_fractions: Union[float, Tuple[float, float]] = None,
        max_attempts: int = 30,
        crop_corner: bool = True,
        module: str = "internal",
    ):
        if center_fractions:
            raise ValueError(f"center_fractions not yet supported for class {type(self)}.")
        if module not in ("internal", "sigpy"):
            raise ValueError("`module` must be one of ('internal', 'sigpy')")
        super().__init__(accelerations)
        if isinstance(calib_size, int):
            calib_size = (calib_size, calib_size)
        self.calib_size = calib_size
        self.max_attempts = max_attempts
        self.crop_corner = crop_corner
        self.module = module

    def __call__(self, out_shape, seed=None, acceleration=None):
        # Design parameters for mask
        nky = out_shape[1]
        nkz = out_shape[2]
        if not acceleration:
            acceleration = self.choose_acceleration()

        # From empirical results, larger dimension should be first
        # for optimal speed.
        if nky < nkz:
            shape = (nkz, nky)
            transpose = True
        else:
            shape = (nky, nkz)
            transpose = False

        # Issue #2: Due to some optimization reasons, using the internal
        # poisson disc module has been faster than using sigpy's
        # default one. In many cases, the call to sigpy hangs.
        module = self.module
        if module == "internal":
            mask = poisson(
                shape,
                acceleration,
                calib=self.calib_size,
                dtype=np.float32,
                seed=seed,
                K=self.max_attempts,
                crop_corner=self.crop_corner,
            )
        elif module == "sigpy":
            mask = sigpy.mri.poisson(
                shape,
                acceleration,
                calib=self.calib_size,
                dtype=np.float32,
                seed=seed,
                max_attempts=self.max_attempts,
                crop_corner=self.crop_corner,
            )
        if transpose:
            mask = mask.transpose()

        # Reshape the mask
        mask = torch.from_numpy(mask.reshape(out_shape))

        return mask

    def get_edge_mask(self, kspace: torch.Tensor, out_shape: Sequence[int] = None):
        """See :method:`MaskFunc.get_edge_mask`.

        Expected `kspace` shape (batch, ky, kz, ...).
        """

        if out_shape is not None and out_shape[1:3] != kspace.shape[1:3]:
            raise ValueError("out_shape must have the same ky and kz dimensions as kspace.")

        if out_shape is None:
            out_shape = kspace.shape
        h, w = kspace.shape[1:3]

        # When crop corner is disabled, all edges are sampled.
        if not self.crop_corner:
            return torch.zeros(out_shape, dtype=torch.float32, device=kspace.device)

        # The edges of the ellipse are not sampled.
        # These points should be set to 1 in the training mask.
        # y: the row dimension, x: the column dimension.
        # Ellipse equation: (x - xc)^2 / w^2 + (y - yc)^2 / h^2 = 1
        grid_y, grid_x = torch.meshgrid(
            torch.arange(h, device=kspace.device, dtype=torch.float32),
            torch.arange(w, device=kspace.device, dtype=torch.float32),
            indexing="ij",
        )
        yc, xc = _get_center(h), _get_center(w)
        outer_ellipse_mask = (grid_y - yc) ** 2 / (h / 2) ** 2 + (grid_x - xc) ** 2 / (
            w / 2
        ) ** 2 > 1

        outer_ellipse_mask = outer_ellipse_mask.type(torch.float32)
        return _reshape_or_tile(outer_ellipse_mask, shape=out_shape, ndim=2)

    def _get_args(self):
        return {
            "accelerations": self.accelerations,
            "calib_size": self.calib_size,
            "max_attempts": self.max_attempts,
            "crop_corner": self.crop_corner,
        }

    def get_str_name(self):
        args = self._get_args()
        return f"{type(self).__name__}-" + "-".join(f"{k}={v}" for k, v in args.items())

    def __str__(self) -> str:
        args = self._get_args()
        args_str = "\n\t" + "\n\t".join(f"{k}={v}" for k, v in args.items()) + "\n\t"
        return f"{type(self)}({args_str})"


@MASK_FUNC_REGISTRY.register()
class RandomMaskFunc1D(MaskFunc):
    """
    RandomMaskFunc creates a sub-sampling mask of a given shape.
    The mask selects a subset of columns from the input k-space data. If the k-space data has N
    columns, the mask picks out:
        1. N_low_freqs = (N * center_fraction) columns in the center corresponding to
           low-frequencies
        2. The other columns are selected uniformly at random with a probability equal to:
           prob = (N / acceleration - N_low_freqs) / (N - N_low_freqs).
    This ensures that the expected number of columns selected is equal to (N / acceleration)
    It is possible to use multiple center_fractions and accelerations, in which case one possible
    (center_fraction, acceleration) is chosen uniformly at random each time the RandomMaskFunc
    object is called.

    For example, if accelerations = [4, 8] and center_fractions = [0.08, 0.04], then there
    is a 50% probability that 4-fold acceleration with 8% center fraction is selected and a 50%
    probability that 8-fold acceleration with 4% center fraction is selected.

    Adapted from https://github.com/facebookresearch/fastMRI/blob/master/common/subsample.py
    """

    def __init__(self, accelerations, calib_size=None, center_fractions=None):
        """
        Args:
            center_fractions (List[float]): Fraction of low-frequency columns to be retained.
                If multiple values are provided, then one of these numbers is chosen uniformly
                each time.
            accelerations (List[int]): Amount of under-sampling. This should have the same length
                as center_fractions. If multiple values are provided, then one of these is chosen
                uniformly each time. An acceleration of 4 retains 25% of the columns, but they may
                not be spaced evenly.
            calib_size (List[int]): Calibration size for scans.
        """
        if not calib_size and not center_fractions:
            raise ValueError("Either calib_size or center_fractions must be specified.")
        if calib_size and center_fractions:
            raise ValueError("Only one of calib_size or center_fractions can be specified")

        self.center_fractions = center_fractions
        self.calib_size = calib_size
        super().__init__(accelerations)

    def __call__(self, shape, seed=None, acceleration=None):
        """
        Args:
            shape (iterable[int]): The shape of the mask to be created. The shape should have
                at least 3 dimensions. Samples are drawn along the second last dimension.
            seed (int, optional): Seed for the random number generator. Setting the seed
                ensures the same mask is generated each time for the same shape.
        Returns:
            torch.Tensor: A mask of the specified shape.
        """
        if len(shape) < 3:
            raise ValueError("Shape should have 3 or more dimensions")

        if seed is not None:
            np_state = np.random.get_state()
        rng = np.random.RandomState(seed) if seed is not None else self.rng

        num_rows = shape[1]
        num_cols = shape[2]
        if self.center_fractions:
            if isinstance(self.center_fractions, Sequence):
                choice = rng.randint(0, len(self.center_fractions))
                center_fraction = self.center_fractions[choice]
            else:
                center_fraction = self.center_fractions
        else:
            center_fraction = self.calib_size / num_cols
        if acceleration is None:
            acceleration = self.choose_acceleration()

        # Create the mask
        num_low_freqs = int(round(num_cols * center_fraction))
        prob = (num_cols / acceleration - num_low_freqs) / (num_cols - num_low_freqs)
        mask = rng.uniform(size=num_cols) < prob
        pad = (num_cols - num_low_freqs + 1) // 2
        mask[pad : pad + num_low_freqs] = True

        # Reshape the mask
        mask_shape = [1 for _ in shape]
        mask_shape[2] = num_cols
        mask = mask.reshape(*mask_shape).astype(np.float32)
        mask = np.concatenate([mask] * num_rows, axis=1)
        mask = torch.from_numpy(mask)

        if seed is not None:
            np.random.set_state(np_state)

        return mask

    def get_edge_mask(self, kspace: torch.Tensor, out_shape: Sequence[int] = None):
        """See :method:`MaskFunc.get_edge_mask`.

        Expected `kspace` shape (batch, ky, kz, ...).
        """
        # TODO: dims should be configured based on the number of dimenions in the input.
        return get_cartesian_edge_mask(kspace, dims=(1, 2), out_shape=out_shape)


def get_cartesian_edge_mask(
    kspace: torch.Tensor,
    dims: Union[int, Sequence[int]],
    out_shape: Sequence[int] = None,
    dtype=torch.float32,
):
    """See :method:`MaskFunc.get_edge_mask`.

    Expected `kspace` shape (batch, ky, kz, ...).
    """
    if isinstance(dims, int):
        dims = (dims,)
    dims = tuple(dims)

    if out_shape is None:
        out_shape = kspace.shape
    if any(out_shape[i] != kspace.shape[i] for i in (0,) + dims):
        raise ValueError(
            "out_shape must have the same shape as kspace along batch and `dims` dimensions."
        )
    if kspace.ndim != len(out_shape):
        raise ValueError("out_shape must have the same number of dimensions as kspace.")

    mask = cplx.get_mask(kspace)
    non_edge_mask = torch.zeros(out_shape, dtype=torch.bool, device=kspace.device)
    batch = kspace.shape[0]

    # Slices for each dimension in the output mask.
    # We will isolate areas in the slices that
    sls = [[slice(None)] * batch for _ in range(kspace.ndim - 1)]
    sls = [list(range(batch))] + sls  # add slice to index batch dimension
    for dim in dims:
        sls[dim] = _get_nonedge_indices_slice(mask, dim=dim)

    # Set the non-edge mask to 1 in the slice region.
    for sl in zip(*sls):
        non_edge_mask[sl] = True
    # Negate the non-edge mask to get the edge mask.
    return (~non_edge_mask).type(dtype)


def _get_nonedge_indices_slice(mask: torch.Tensor, dim: int) -> List[slice]:
    """Get the slice of indices that correspond to non-edge region along the given dimension.

    Args:
        mask: A mask of shape (batch, ...).
        dim: The dimension along which to get the slice.
            The nonedge slice cannot be computed along the batch dimension (i.e. dim>0).

    Return:
        List[slice]: The slice of indices that correspond to non-edge region
            for each example in the batch.
    """
    # If dim is negative, convert it to the positive index.
    if dim < 0:
        dim = mask.ndim + dim
    reduce_idxs = [i for i in range(1, mask.ndim) if i != dim]
    is_zero = mask.sum(reduce_idxs) == 0  # shape: (batch, dim)
    is_zero = F.pad(is_zero, (1, 1), value=True)

    sls = []
    # For each example in the batch,
    # determine the first and last non-zero indices.
    for batch_idx in range(is_zero.shape[0]):
        zero_offset = torch.where(~(is_zero[batch_idx, :-1] & is_zero[batch_idx, 1:]))[0]
        i0 = zero_offset[0].cpu().item()
        i1 = zero_offset[-1].cpu().item()
        sls.append(slice(i0, i1))
    return sls


class MaskLoader(MaskFunc):
    """Loads masks from predefined file format instead of computing on the fly."""

    def __init__(self, accelerations, masks_path, mask_type: str = "poisson", mode="train"):
        assert isinstance(accelerations, (int, float)) or len(accelerations) == 1
        assert mode in ["train", "eval"]
        if isinstance(accelerations, (int, float)):
            accelerations = (accelerations,)
        super().__init__(accelerations)

        accel = float(self.accelerations[0])
        self.train_masks = None
        self.eval_data = torch.load(os.path.join(masks_path, f"{mask_type}_{accel}x_eval.pt"))
        if mode == "train":
            self.train_masks = np.load(os.path.join(masks_path, f"{mask_type}_{accel}x.npy"))

    def __call__(self, out_shape, seed=None, acceleration=None):
        if acceleration is not None and acceleration not in self.accelerations:
            raise RuntimeError(
                "MaskLoader.__call__ does not currently support ``acceleration`` argument"
            )

        if seed is None:
            # Randomly select from the masks we have
            idx = np.random.choice(len(self.train_masks))
            mask = self.train_masks[idx]
        else:
            data = self.eval_data
            masks = self.eval_data["masks"]
            mask = masks[data["seeds"].index(seed)]

        mask = mask.reshape(out_shape)
        return torch.from_numpy(mask)


def _reshape_or_tile(x: torch.Tensor, shape: Sequence[int], ndim: int) -> torch.Tensor:
    """Reshape the tensor to the output shape or k-space shape.

    k-space shape will require tiling the tensor to match the
    k-space shape.
    """
    leading_dims = ndim + 1  # number of spatial dimensions + batch dimension

    # This is a broadcasted output shape, where all non-zero dimensions match
    # the shape of the tensor and all other dimensions are 1.
    shape_tensor = torch.as_tensor(shape)
    if shape[:leading_dims] == (1, *x.shape) and all(shape_tensor[leading_dims:] == 1):
        return x.reshape(shape)

    leading_dims = ndim + 1  # number of spatial dimensions + batch dimension
    extra_dims = len(shape) - (leading_dims)
    mask_shape = (1,) + shape[1:leading_dims] + (1,) * extra_dims
    x = x.reshape(mask_shape)
    return x.repeat(shape[0], 1, 1, *shape[leading_dims:])


def _get_center(size: int) -> float:
    """Get the center of the matrix."""
    return size // 2 if size % 2 == 1 else size // 2 - 0.5


# ================================================================ #
# Adapted from sigpy.
# Duplicated because of https://github.com/mikgroup/sigpy/issues/54
# TODO: Remove once https://github.com/mikgroup/sigpy/issues/54 is
# solved and added to release.
# ================================================================ #
def poisson(
    img_shape,
    accel,
    K=30,
    calib=(0, 0),
    dtype=np.complex,
    crop_corner=True,
    return_density=False,
    seed=0,
):
    """Generate Poisson-disc sampling pattern

    Args:
        img_shape (tuple of ints): length-2 image shape.
        accel (float): Target acceleration factor. Greater than 1.
        K (float): maximum number of samples to reject.
        calib (tuple of ints): length-2 calibration shape.
        dtype (Dtype): data type.
        crop_corner (bool): Toggle whether to crop sampling corners.
        return_density (bool): Toggle whether to return sampling density.
        seed (int): Random seed.

    Returns:
        array: Poisson-disc sampling mask.

    References:
        Bridson, Robert. "Fast Poisson disk sampling in arbitrary dimensions."
        SIGGRAPH sketches. 2007.

    """
    y, x = np.mgrid[: img_shape[-2], : img_shape[-1]]
    x = np.maximum(abs(x - img_shape[-1] / 2) - calib[-1] / 2, 0)
    x /= x.max()
    y = np.maximum(abs(y - img_shape[-2] / 2) - calib[-2] / 2, 0)
    y /= y.max()
    r = np.sqrt(x**2 + y**2)

    slope_max = 40
    slope_min = 0
    if seed is not None:
        rand_state = np.random.get_state()
    else:
        seed = -1  # numba does not play nicely with None types
    while slope_min < slope_max:
        slope = (slope_max + slope_min) / 2.0
        R = 1.0 + r * slope
        mask = _poisson(img_shape[-1], img_shape[-2], K, R, calib, seed)
        if crop_corner:
            mask *= r < 1

        est_accel = img_shape[-1] * img_shape[-2] / np.sum(mask[:])

        if abs(est_accel - accel) < 0.1:
            break
        if est_accel < accel:
            slope_min = slope
        else:
            slope_max = slope

    if seed is not None and seed > 0:
        np.random.set_state(rand_state)
    mask = mask.reshape(img_shape).astype(dtype)
    if return_density:
        return mask, r
    else:
        return mask


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _poisson(nx, ny, K, R, calib, seed=None):

    mask = np.zeros((ny, nx))
    f = ny / nx

    if seed is not None and seed > 0:
        np.random.seed(int(seed))

    pxs = np.empty(nx * ny, np.int32)
    pys = np.empty(nx * ny, np.int32)
    pxs[0] = np.random.randint(0, nx)
    pys[0] = np.random.randint(0, ny)
    m = 1
    while m > 0:

        i = np.random.randint(0, m)
        px = pxs[i]
        py = pys[i]
        rad = R[py, px]

        # Attempt to generate point
        done = False
        k = 0
        while not done and k < K:

            # Generate point randomly from R and 2R
            rd = rad * (np.random.random() * 3 + 1) ** 0.5
            t = 2 * np.pi * np.random.random()
            qx = px + rd * np.cos(t)
            qy = py + rd * f * np.sin(t)

            # Reject if outside grid or close to other points
            if qx >= 0 and qx < nx and qy >= 0 and qy < ny:

                startx = max(int(qx - rad), 0)
                endx = min(int(qx + rad + 1), nx)
                starty = max(int(qy - rad * f), 0)
                endy = min(int(qy + rad * f + 1), ny)

                done = True
                for x in range(startx, endx):
                    for y in range(starty, endy):
                        if mask[y, x] == 1 and (
                            ((qx - x) / R[y, x]) ** 2 + ((qy - y) / (R[y, x] * f)) ** 2 < 1
                        ):
                            done = False
                            break

                    if not done:
                        break

            k += 1

        # Add point if done else remove active
        if done:
            pxs[m] = qx
            pys[m] = qy
            mask[int(qy), int(qx)] = 1
            m += 1
        else:
            pxs[i] = pxs[m - 1]
            pys[i] = pys[m - 1]
            m -= 1

    # Add calibration region
    mask[
        int(ny / 2 - calib[-2] / 2) : int(ny / 2 + calib[-2] / 2),
        int(nx / 2 - calib[-1] / 2) : int(nx / 2 + calib[-1] / 2),
    ] = 1

    return mask
